% [n1,p] = gibbs_sampler(n0,alpha,beta,W,layer1,layer2,Nsteps,Mm,Mn)
%
% Implementation of the harmonium gibbs sampler where 
% n0		are the starting points of the chains
% alpha,beta,W 	are parameters of the model
% layer1	either 'poisson' or 'binom'
% layer2	only 'binom' at this point
% Nsteps	#steps chains should run
% Mm,Mn		#of states in hidden/observerd layer
%
% Peter Gehler
function [n1,p] = gibbs_sampler(n0,alpha,beta,W,layer1,layer2,Nsteps,Mm,Mn)

if nargin < 9,Mn = 1;end
if nargin < 7,Nsteps = 1;end
[D,N] = size(n0);
if size(beta,2) ~= N, beta = repmat(beta,1,N);end
if size(alpha,2) ~= N, alpha = repmat(alpha,1,N); end

switch(strcat(layer1,layer2));
 case 'poissonbinom'
  cnt = Nsteps;
  n1 = n0;
  while cnt
      cnt = cnt-1;
      p = sigmoid(beta+W*n1);
      % Statistics toolbox variant
      %m0 = binornd(Mm,p);
      %n1 = sparse(poissrnd(exp(alpha+W'*m0)));
      % lightspeed variant
      m0 = randbinom(p,Mm);
      n1 = sparse(my_poissrnd(exp(alpha+W'*m0)));
      
  end
 case 'binombinom' %%%%%% BINOM / BINOM %%%%%%%%%%
  cnt = Nsteps;
  n1 = n0;
  while cnt
      cnt = cnt-1;
      p = sigmoid(beta+W*n1);
      m0 = binornd(Mm,p);
      q = sigmoid(alpha+W'*m0);
      n1 = sparse(binornd(Mn,q));
  end
  
  %[n1,p] = sample_bern_bern(n0,alpha,beta,W,Mm,Mn,Nsteps);
 otherwise
  error('no such method');
end


return;

