
#include "mex.h"
#include <math.h>
/*
NumNonZeros.C
% NUMNONZEROS Faster alternative for nnz
% Returns the number of potenially non-zero elements in a sparse matrix
% or the exact number of non-zeros in a full matrix
%
% much faster than the built-in nnz
*/

void mexFunction(
    int nlhs, mxArray *plhs[],
    int nrhs, const mxArray *prhs[])
{
   /* Error checking on inputs */
   if (nrhs<1) mexErrMsgTxt("Not enough input arguments.");
   if (nrhs>1) mexErrMsgTxt("Too many input arguments.");
   if (nlhs>1) mexErrMsgTxt("Too many output arguments.");
   plhs[0]=mxCreateDoubleMatrix(1,1,mxREAL);
   if (mxIsSparse(prhs[0]))
     *mxGetPr(plhs[0])=*(mxGetJc(prhs[0])+mxGetN(prhs[0]));
   else if (mxIsDouble(prhs[0])){
     int s=0, i, n;
     double *xr, *xi;
     n=mxGetNumberOfElements(prhs[0]);
     xr=mxGetPr(prhs[0]);
     xi=mxGetPi(prhs[0]);
     if (xi!=NULL)
       for (i=0;i<n;i++){
         if (xr[i]!=0) s++;
         else if (xi[i]!=0) s++;
       }
     else
       for (i=0;i<n;i++) if (xr[i]!=0) s++;
     *mxGetPr(plhs[0])=s;
   }
   else
     mexErrMsgTxt("Function not defined for variables of input class");
}
