#include <math.h>
#include "mex.h"
#include "matrix.h"


#define MALLOC malloc
#define FREE free

/* the angles between one test document and all the training documents. 
Needs to be global in order to sort according to it an index */
double *angle;

/*
  compute S .* (a*b') where S is sparse and a,b is outer product 
  with matching dimensions
*/
int compare (const void * a, const void * b)
{
    /*
      <0	*elem1 goes before *elem2
      0	*elem1 == *elem2
      >0	*elem1 goes after *elem2
    */
        return (angle[ *(int*)a] < angle[*(int*)b] ? 1:
	    angle[ *(int*)a] > angle[*(int*)b] ?  -1: 0);
    
	/*printf("a = %d\n", *(int*)a);
      return( *(double*)a < *(double*)b ? -1 :
      *(double*)a > *(double*)b ?  1 : 0);*/
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[]) { 
    double *trainDocs, *testDocs,*precision, *recall, norm;;
    int i,j,tt, numCorrect,correctClass,counter,ii, test_indx, train_indx, next_test_indx,next_train_indx;
    unsigned int startIndTest,startIndTrain, *index;
    int *ytrain,*ytest, *jc_train, *jc_test, *ir_train, *ir_test;
    int *evalPoints, nEvalPoints, nextEvalPoint, nClasses, *sizeClass;
    double* confusionTable;
    unsigned int nTopics,nDocs,nTestDocs;
    int MULTICLASS = 0;

    if ((nrhs!=4)||(!((nlhs==2)||(nlhs==3)))){
	printf("usage: [Rec,Prec] = mex_rpc(Ytrain,Ytest,ytrain,ytest)\n");
	printf("   or: [Rec,Prec,T] = mex_rpc(Ytrain,Ytest,ytrain,ytest)\n");
	printf("   where ytrain,ytest have to be int32\n");
	printf("       ytrain and ytest are either of size [nDocs,1] or [nDocs,nClasses]\n");
	printf("       Rec is Recall, Prec, Precsion and T a confusion table\n");
	return;
    }

    nDocs = mxGetN(prhs[0]);
    nTopics = mxGetM(prhs[0]);

    nTestDocs = mxGetN(prhs[1]);

    testDocs = mxGetPr(prhs[1]);
    trainDocs = mxGetPr(prhs[0]);

    ytrain = (int*)mxGetPr(prhs[2]);
    ytest = (int*)mxGetPr(prhs[3]);

    if (mxGetN(prhs[2]) > 1){
	MULTICLASS = 1;
	printf("each data point might have more than one label\n");
    }


    /*
      printf("nDocs = %d\nNTopics = %d\n",nDocs,nTopics);
      printf("nTestDocs = %d\n",nTestDocs);
    */

    angle = mxGetPr(mxCreateDoubleMatrix(nDocs,1,mxREAL)); /*array for similarities */
    index = (int *)malloc(nDocs * sizeof(int));

    /* 
       build an array of 'evaluation' points, that is the number of retrieved documents. i.e
       1,2,4,8,16,...,nDocs
    */
    nEvalPoints = ceil(log2(nDocs))+1;
    evalPoints = (int*)malloc(nEvalPoints* sizeof(int));
    /*
      printf("nEvalPoints = %d\n",nEvalPoints);
    */
    evalPoints[0] = 1;
    for (tt=1;tt<nEvalPoints-1;tt++){
	evalPoints[tt] = 2*evalPoints[tt-1];
    }
    evalPoints[nEvalPoints-1] = nDocs;

    
    printf("evaluation points : ");
    for (tt=0;tt<nEvalPoints;tt++)
	printf("%d  ",evalPoints[tt]);
    printf("\n");

    /* create output matrices */
    plhs[0] = mxCreateDoubleMatrix(nEvalPoints,1,mxREAL);
    plhs[1] = mxCreateDoubleMatrix(nEvalPoints,1,mxREAL);

    recall = (double*)mxGetPr(plhs[0]);
    precision = (double*)mxGetPr(plhs[1]);

    for (tt=0;tt<nEvalPoints;tt++){
	precision[tt] = 0;
	recall[tt] = 0;
    }


    if (!(MULTICLASS)){
	/*    printf("nClasses = %d\n",nClasses);*/
	nClasses = maximum(ytrain,nDocs) + 1;
	sizeClass = (int*)calloc(nClasses, sizeof(int));
	for (tt=0;tt<nDocs;tt++)
	    sizeClass[ytrain[tt]]++;
	for (tt=0;tt<nClasses;tt++)
	    printf("class %d size %d\n",tt,sizeClass[tt]);
    }
    else {
	nClasses = mxGetN(prhs[2]);
    }

    if (nlhs==3){
	plhs[2] = mxCreateDoubleMatrix(nClasses,nClasses,mxREAL);
	confusionTable = mxGetPr(plhs[2]);
	for (tt=0;tt<nClasses*nClasses;tt++)
	    confusionTable[tt] = 0;
    }



    /*
      normalize the training and testing documents
      trainDocs = trainDocs./repmat(sqrt(sum(trainDocs.*trainDocs)),nTopics,1);
    */
    if (mxIsSparse(prhs[0])){
	int next_indx,old_indx, indx;
	indx = 0;
	jc_train = mxGetJc(prhs[0]);
	ir_train = mxGetIr(prhs[0]);
	for (tt = 0; tt < nDocs; tt++) {
	    next_indx = jc_train[tt+1]; /* next column index */
	    old_indx = indx;
	    norm = 0;
	    while (indx < next_indx){
		norm += trainDocs[indx]*trainDocs[indx];
		indx++;
	    }
	    norm = sqrt(norm);
	    indx = old_indx;
	    while (indx < next_indx){
		trainDocs[indx] /= norm;
		indx++;
	    }
	}
    }
    else{
	for (tt=0;tt<nDocs;tt++){
	    norm = 0;
	    startIndTrain = tt * nTopics;
	    for (i=0;i<nTopics;i++)
		norm += trainDocs[startIndTrain+i]*trainDocs[startIndTrain+i];
	    norm = sqrt(norm);
	    for (i=0;i<nTopics;i++)
		trainDocs[startIndTrain+i] /= norm;
	}
    }
    
    if (mxIsSparse(prhs[1])){
	int next_indx,old_indx,indx;
	indx = 0;
	jc_test = mxGetJc(prhs[1]);
	ir_test = mxGetIr(prhs[1]);
	for (tt = 0; tt < nTestDocs; tt++) {
	    next_indx = jc_test[tt+1]; /* next column index */
	    old_indx = indx;
	    norm = 0;
	    while (indx < next_indx){
		norm += testDocs[indx]*testDocs[indx];
		indx++;
	    }
	    norm = sqrt(norm);
	    indx = old_indx;
	    while (indx < next_indx){
		testDocs[indx] /= norm;
		indx++;
	    }
	}
    }
    else{
	for (tt=0;tt<nTestDocs;tt++){
	    norm = 0;
	    startIndTest = tt * nTopics;
	    for (i=0;i<nTopics;i++)
		norm += testDocs[startIndTest+i]*testDocs[startIndTest+i];
	    norm = sqrt(norm);
	    for (i=0;i<nTopics;i++)
		testDocs[startIndTest+i] /= norm;
	}
    }
    
    /*    for (tt=0;tt<nClasses;tt++)
	  printf("elems in class %d = %d\n",tt,sizeClass[tt]);
    */

    printf("starting retrieval\n");
    /* for each test document */    


    /* for sparse input */
    if (MULTICLASS){ /* Multiclass version */
	
	for (i=0;i<nTestDocs;i++){
	    int jj,kk,sizeSubset = 0;
	    
	    if (mxIsSparse(prhs[0])){
		next_test_indx = jc_test[i+1];
		for (j=0;j<nDocs;j++){
	
		    test_indx = jc_test[i];
		    train_indx = jc_train[j];
		    next_train_indx = jc_train[j+1];
		    angle[j] = 0;
		    index[j] = j;
		    while (train_indx < next_train_indx){
			/* reached the end, reset the test index */
			if (test_indx == next_test_indx){ 
			    break;
			}
			/* element in test but not in train */
			if (ir_train[train_indx] < ir_test[test_indx]){
			    train_indx++;
			    continue;
			}
			/* element in train but not in test */
			if (ir_test[test_indx] < ir_train[train_indx]){
			    test_indx++;
			    continue;
			}
			angle[j] += testDocs[test_indx] * trainDocs[train_indx];
			test_indx++;
			train_indx++;
		    }
		}

	    }
	    else{
		startIndTest = i * nTopics;
		for (j=0;j<nDocs;j++){
		    startIndTrain = j * nTopics;
		    angle[j] = 0;
		    index[j] = j;
		    /* taking the dot product of test and train vector */
		    for (tt=0;tt<nTopics;tt++)
			angle[j] += testDocs[startIndTest+tt] * trainDocs[startIndTrain+tt];
		    
		    angle[j] = angle[j];
		}
	    }
	    qsort(index,nDocs,sizeof(int),compare);
	
	    nextEvalPoint = evalPoints[0];
	    counter = 0;
	    numCorrect = 0;

	    /* compute the size of the subset which is relevant to this 
	       query */
	    for (jj=0;jj<nDocs;jj++){ /* for each document */
		for (kk=0;kk<nClasses;kk++){ /* check if at least one class is equal
						to one test class */
		    if ((ytrain[jj+kk*nDocs] > 0) && (ytest[i+kk*nTestDocs] > 0)){
			sizeSubset++;
			continue;
		    }
		}
	    }
	    

	    for (tt=0;tt<nDocs;tt++){
		/*  match of query class and retrieved class  */
		
		for (ii=0;ii<nClasses;ii++){
		    if ((ytrain[index[tt]+ii*nDocs]>0) && (ytest[i+ii*nTestDocs]>0)){
			numCorrect++;
			break;
		    }
		}
		
		if (tt==nextEvalPoint-1){
		    /*
		       update recall and precision values
		    */
		    precision[counter]  += (double)numCorrect/(double)(evalPoints[counter]);
		    recall[counter] += (double)numCorrect/(double)(sizeSubset);

		    /*
		      jump to next evaluation point
		    */
		    counter++;
		    nextEvalPoint = evalPoints[counter];
		}
	    }
	}
    } /* Multiclass version */
    else {
	for (i=0;i<nTestDocs;i++){

	    if (mxIsSparse(prhs[0])){
		next_test_indx = jc_test[i+1];
		
		for (j=0;j<nDocs;j++){
		    test_indx = jc_test[i];
		    train_indx = jc_train[j];
		    next_train_indx = jc_train[j+1];
		    angle[j] = 0;
		    index[j] = j;
		    while (train_indx < next_train_indx){
			/* reached the end, reset the test index */
			if (test_indx == next_test_indx){ 
			    break;
			}
			/* element in test but not in train */
			if (ir_train[train_indx] < ir_test[test_indx]){
			    train_indx++;
			    continue;
			}
			/* element in train but not in test */
			if (ir_test[test_indx] < ir_train[train_indx]){
			    test_indx++;
			    continue;
			}
			angle[j] += testDocs[test_indx] * trainDocs[train_indx];
			test_indx++;
			train_indx++;
		    }
		    
		}
	    }

	    else {
		/* start index of the test document */
		startIndTest = i * nTopics;
		
		for (j=0;j<nDocs;j++){
		    startIndTrain = j * nTopics;
		    angle[j] = 0;
		    index[j] = j;

		    /* taking the dot product of test and train vector */
		    for (tt=0;tt<nTopics;tt++)
			angle[j] += testDocs[startIndTest+tt] * trainDocs[startIndTrain+tt];
		    
		    angle[j] = angle[j];
		}
	    }
	    qsort(index,nDocs,sizeof(int),compare);
	
	    /*	
	      for (tt=0;tt<nDocs;tt++){
	      printf("index(%d) = %d\n",tt, (int)ytrain[index[tt]]);
	      }
	    */

	    correctClass = ytest[i];
	
	    nextEvalPoint = evalPoints[0];
	    counter = 0;
	    numCorrect = 0;

	    for (tt=0;tt<nDocs;tt++){
		/*
		  match of query class and retrieved class
		*/
		if (ytrain[index[tt]]==correctClass)
		    numCorrect++;

		if ((nlhs==3) && (tt==0))
		    confusionTable[ytrain[index[tt]]+correctClass*nClasses]++;

		if (tt==nextEvalPoint-1){
		    /* 
		       update recall and precision values
		    */
		
		
		    /*		printf("counter = %d\n",counter);
				printf("numCorrect = %d\n",numCorrect);
				printf("evalPoints[counter] = %d\n",evalPoints[counter]);
				printf("correctClass = %d\n",correctClass);
		    */
		    precision[counter]  += (double)numCorrect/(double)(evalPoints[counter]);

		    recall[counter] += (double)numCorrect/(double)(sizeClass[correctClass]);
		
		    /*
		      jump to next evaluation point
		    */
		    counter++;
		    nextEvalPoint = evalPoints[counter];
		}
	    }
	}
	
    }
    /* average the recall and precision value */
    for (tt=0;tt<nEvalPoints;tt++){	
	precision[tt] /= nTestDocs;
	recall[tt] /= nTestDocs;
    }
    
    free(index);
    free(evalPoints);

    
    return;
}


int maximum(int* a,int nelem){
    int t, max;
    
    if (nelem == 1) 
	return a[0];
    else
	max = a[0];

    for (t=1;t<nelem;t++){
	if (a[t] > max) max = a[t];
    }

    return max;

}
