#include <math.h>
#include "mex.h"

/* these 2 #define lines help make the later code more readable */
/* Input Arguments */
#define PARAMETER_IN    prhs[0]

/* Output Arguments */
#define RESULT_OUT  plhs[0]

/*#define LITTLE_ENDIAN  0*/


#define POW_2_20 (1 << 20)
#define POW_1023 (1023*POW_2_20)
#define EXP_A2 (POW_2_20/(2*M_LN2))
/* #define EXP_A2 756388 // for integer y*/

static union 
{
  double d;
    struct {
#ifdef LITTLE_ENDIAN
    int j,i;
#else 
    int i,j;
#endif
    } n;
} _eco, _eco2;

#define EXP2(y) (_eco2.n.i = POW_1023 - EXP_A2*(y), \
    _eco.n.i = 2*POW_1023 - _eco2.n.i, _eco.d/_eco2.d) 


#define EXP_A (1048576/0.69314718055994530942)
#define EXP_C 60801
#define EXP(y) (_eco.n.i = EXP_A*(y) + (1072693248 - EXP_C), _eco.d)

void mexexp(double*y, double*yp, int m) {
  while(m>0) {
    m--;
    yp[m]=EXP2(y[m]);
  }
}


void mexFunction( int nlhs, mxArray *plhs[], 
                  int nrhs, const mxArray *prhs[] )
{ 
    double *yp; 
    double *y; 
    unsigned int m,n; 
    int i,j;
    int nel;
    int *dims;

    /* Check for proper number of arguments */    
        if (nrhs != 1) { 
        mexErrMsgTxt("One input argument required."); 
    } else if (nlhs > 1) {
        mexErrMsgTxt("Too many output arguments."); 
    } 
    
    m = mxGetM(PARAMETER_IN); 
    n = mxGetN(PARAMETER_IN);

    /* Create a matrix for the return argument */ 
    /*RESULT_OUT = mxCreateDoubleMatrix(m, n, mxREAL); */

    
    dims= (int*)mxMalloc(2*sizeof(int));
    dims[0] = m;
    dims[1] = n;
    
    RESULT_OUT = mxCreateDoubleMatrix(0,0,mxREAL);
    mxSetDimensions(RESULT_OUT,dims,2);
    /*nel = mxGetNumberOfElements(RESULT_OUT);*/
    mxSetPr(RESULT_OUT,mxMalloc(m*n*sizeof(double)));
    
    /* Assign pointers to the various parameters */ 
    yp = mxGetPr(RESULT_OUT);    
    y = mxGetPr(PARAMETER_IN);

    /* Do the actual computation*/
    if(m==1)
        mexexp(y,yp,n); 
    else{
        for(i=0;i<n;i++){
            for(j=0;j<m;j++){
                yp[(i*m)+j]=EXP2(y[(i*m)+j]);
            }
        }
     }
    return;
}
