function r = my_poissrnd(lambda,varargin)
%POISSRND Random arras from the Poisson distribution.
%
% same as the Poissrnd function from the statistics toolbox but
% this function uses randgamma and randbinom from the lightspeed
% toolbox and therefore no statistics licence is needed to run this function!
%
% Peter Gehler

if nargin < 1
    error('stats:poissrnd:TooFewInputs','Requires at least one input argument.');
end

%[err, sizeOut, numelOut] = statsizechk(1,lambda,varargin{:});
%if err > 0
%    error('stats:poissrnd:InputSizeMismatch','Size information is inconsistent.');
%end

sizeOut = size(lambda);

lambda(lambda < 0) = NaN;
%if isscalar(lambda)
%    lambda = repmat(lambda, numelOut, 1);
%else
lambda = lambda(:);
%end

%Initialize r to zero.
if isa(lambda,'single')
   r = zeros(sizeOut,'single');
else
   r = zeros(sizeOut);
end

r(isinf(lambda)) = Inf;

% For large lambda, use the method of Ahrens and Dieter as
% described in Knuth, Volume 2, 1998 edition.
k = find(15 <= lambda & lambda < Inf);
if ~isempty(k)
   alpha = 7/8;
   lk = lambda(k);
   m = floor(alpha * lk);

   % Generate m waiting times, all at once
   %x = randg(m);
   x = randgamma(m); % lightspeed toolbox
   t = (x <= lk);

   % If we did not overshoot, then the number of additional times
   % has a Poisson distribution with a smaller mean.
   r(k(t)) = m(t) + my_poissrnd(lk(t)-x(t));

   % If we did overshoot, then the times up to m-1 are uniformly
   % distributed on the interval to x, so the count of times less
   % than lambda has a binomial distribution.
   % r(k(~t)) = binornd(m(~t)-1, lk(~t)./x(~t));
   r(k(~t)) = randbinom(lk(~t)./x(~t), m(~t)-1);  % from the lightspeed
end

% For small lambda, generate and count waiting times.
j = find(lambda < 15);
p = zeros(numel(j),1);
while ~isempty(j)
    p = p - log(rand(numel(j),1));
    t = (p < lambda(j));
    j = j(t);
    p = p(t);
    r(j) = r(j) + 1;
end

% Return NaN if LAMBDA is negative.
r(isnan(lambda)) = NaN;
